/** @file

    @copy Siemens Schweiz AG, vertraulich
    
    $Id: integer.hxx 169 2006-12-15 15:26:48Z chawama0 $
    */
//       1          2         3         4         5         6         7         8
//  56789012345676890123456789012345678901234567890123456789012345678901234567890

#ifndef __IMAGE_INTEGER__
#define __IMAGE_INTEGER__

#include <image/type.hxx>
#include <image/traits.hxx>

namespace image {
  //! @addtogroup imageBasetypes
  //@{

  //=============================================================================
  namespace byteorder {
    /*! @name ByteOrder
        Byte-Folge herausfinden zur Kompilierzeit. */
    //@{
    namespace { // Hilfsvariabeln zur Bestimmung der System-Byte-Folge
      static const unsigned int AUX((222<<8) + 111);
      struct AUX_MAP{unsigned char first, second;};
    }
    const bool change //!< System hat H&ouml;chstes-Byte-Zuerst
        (reinterpret_cast<const AUX_MAP*>(&AUX)->first==222
         && reinterpret_cast<const AUX_MAP*>(&AUX)->second==111);
    const bool unchanged //!< System has Tiestes-Byte-Zuerst
        (reinterpret_cast<const AUX_MAP*>(&AUX)->first==111
         && reinterpret_cast<const AUX_MAP*>(&AUX)->second==222);
    //@}
  }

  //=============================================================================
  namespace {
    // Hilfsklasse zur Bestimmung des passenden vorzeichenlosen Integer-Typs
    template<int ENUM> class UIntTypes {};
    template<> class UIntTypes<1> {
      public: typedef unsigned char Type;
    };
    template<> class UIntTypes<2> {
      public: typedef unsigned short Type;
    };
    template<> class UIntTypes<3> {
      public: typedef unsigned int Type;
    };
    template<> class UIntTypes<4> {
      public: typedef unsigned long Type;
    };
  }

  //=============================================================================
  //! Integer finden, in den eine vorgegebene Anzahl Bytes hinein passt.
  /*! Die Benutung dieser Klasse erfolgt so:

        -# Der Code <code>image::Int<3>::type i;</code> definiert eine
           Variabel @c i, die gerade gross genug ist, um drei
           Bytes aufzunehmen.
        -# Normalerweise werden im Image aber gleich die Typen Int1, Int2,
           Int3 und Int4 verwendet.
      */
  template<int BYTES, class TRAITS=BaseTraits>
      class Int: public BaseType {

          //............................................................ typedefs
        public:
          
          // Suche die richtige Vorlage der Klasse UIntTypes, in deren
          // Type-Groesse unsere Anzahl BYTES gerade hinein passt. Suche
          // vom Kleinsten zum Groessten. Bei nicht passenden Groessen
          // wird der naechstgroesste Typ genommen, z.B. Int<3>::Type
          // sollte einen Vier-Byte-Integer ergeben (groesser oder gleich
          // drei Byte).
          typedef typename UIntTypes<
            BYTES>=sizeof(UIntTypes<1>::Type) ? 1
            : BYTES>=sizeof(UIntTypes<2>::Type) ? 2
            : BYTES>=sizeof(UIntTypes<3>::Type) ? 3
            : BYTES>=sizeof(UIntTypes<4>::Type) ? 4
            : -1 // Fehler!
            >::Type Type;

          //............................................................. methods
        public:

          Int() {_data.integer=0;}

          bool check() const {return TRAITS(*this).check();}

          Int* alloc() const {return new Int;}

          friend class boost::serialization::access;
          /*! @todo Byte-Order ber&uuml;cksichtigen. */
          template<class ARCHIVE>
              void serialize(ARCHIVE& ar, const unsigned int version) {
            //ar & boost::serialization::base_object<BaseType>(*this);
            // Oder retour (umgekehrte Folge, siehe byteorder::change)
            // oder Start bei sizeof(Type)-Bytes statt bei 0 (und
            // entsprechend bis sizeof(Type))
            // -> Bei Gelegenheit nachdenken (oder experimentieren) und testen!
            for (int i=0; i<BYTES; ++i)
              ar & _data.bytes[i];
          }

          operator unsigned long() const {
            return _data.integer;
          }

          Int& operator=(unsigned long i) {
            _data.integer = i;
            return *this;
          }

          //........................................................... variables
        private:

          union {
              char bytes[BYTES];
              Type integer;
          } _data;
      };

  //=============================================================================
  typedef Int<1> Int1; //!< vorzeichenloser Ein-Byte-Integer
  typedef Int<2> Int2; //!< vorzeichenloser Zwei-Byte-Integer
  typedef Int<3> Int3; //!< vorzeichenloser Drei-Byte-Integer
  typedef Int<4> Int4; //!< vorzeichenloser Vier-Byte-Integer


  //=============================================================================
  namespace {
    // Hilfsklasse zur Bestimmung des passenden vorzeichenbehafteten Integer-Typs
    template<int ENUM> class SIntTypes {};
    template<> class SIntTypes<1> {
      public: typedef unsigned char Type;
    };
    template<> class SIntTypes<2> {
      public: typedef unsigned short Type;
    };
    template<> class SIntTypes<3> {
      public: typedef unsigned int Type;
    };
    template<> class SIntTypes<4> {
      public: typedef unsigned long Type;
    };
  }

  //=============================================================================
  //! Integer finden, in den eine vorgegebene Anzahl Bytes hinein passt.
  /*! Die Benutung dieser Klasse erfolgt so:

        -# Der Code <code>image::Sint<3>::type i;</code> definiert eine
           Variabel @c i, die gerade gross genug ist, um drei
           Bytes aufzunehmen.
        -# Normalerweise werden im Image aber gleich die Typen Sint1, Sint2,
           Sint3 und Sint4 verwendet.
      
      @note Es wird stillschweigend vorausgesetzt, dass alle Integer
            im Zweierkomplement codiert sind. */
  template<int BYTES, class TRAITS=BaseTraits>
      class SInt: public BaseType {

          //............................................................ typedefs
        public:
          // Suche die richtige Vorlage der Klasse SIntTypes, in deren
          // Type-Groesse unsere Anzahl BYTES gerade hinein passt. Suche
          // vom Kleinsten zum Groessten. Bei nicht passenden Groessen
          // wird der naechstgroesste Typ genommen, z.B. SInt<3>::Type
          // sollte einen Vier-Byte-Integer ergeben (groesser oder gleich
          // drei Byte).
          typedef typename SIntTypes<
            BYTES>=sizeof(SIntTypes<1>::Type) ? 1
            : BYTES>=sizeof(SIntTypes<2>::Type) ? 2
            : BYTES>=sizeof(SIntTypes<3>::Type) ? 3
            : BYTES>=sizeof(SIntTypes<4>::Type) ? 4
            : -1 // Fehler!
            >::Type Type;

          //............................................................. methods
        public:

          SInt() {_data.integer=0;}

          bool check() const {return TRAITS(*this).check();}

          SInt* alloc() const {return new SInt;}

          friend class boost::serialization::access;
          /*! @todo Byte-Order ber&uuml;cksichtigen. */
          template<class ARCHIVE>
              void serialize(ARCHIVE& ar, const unsigned int version) {
            //ar & boost::serialization::base_object<BaseType>(*this);
            // Oder retour (umgekehrte Folge, siehe byteorder::change)
            // oder Start bei sizeof(Type)-Bytes statt bei 0 (und
            // entsprechend bis sizeof(Type))
            // -> Bei Gelegenheit nachdenken (oder experimentieren) und testen!
            for (int i=0; i<BYTES; ++i)
              ar & _data.bytes[i];
          }

          operator unsigned long() const {
            return _data.integer;
          }

          SInt& operator=(unsigned long i) {
            _data.integer = i;
            return *this;
          }

          //........................................................... variables
        private:

          union {
              char bytes[BYTES];
              Type integer;
          } _data;

      };

  //=============================================================================
  typedef SInt<1> SInt1; //!< vorzeichenbehafteter Ein-Byte-Integer
  typedef SInt<2> SInt2; //!< vorzeichenbehafteter Zwei-Byte-Integer
  typedef SInt<3> SInt3; //!< vorzeichenbehafteter Drei-Byte-Integer
  typedef SInt<4> SInt4; //!< vorzeichenbehafteter Vier-Byte-Integer

  //@}
}

#endif
